PROGRAM Protpars (infile, treefile, OUTPUT);
  (* version 3.1. (c) Copyright 1986, 1987, 1988 by the University of
     Washington and Joseph Felsenstein.  Written by Joseph Felsenstein.
     Permission is granted to copy and use this program provided no
     fee is charged for it and provided that this copyright
     notice is not removed. *)

CONST maxsp = 50;    (* maximum number of species               *)
   maxsz = 100;      (* size of pointer array.  >= 2*maxsp - 1  *)
   maxsites = 200;   (* max. number amino acids per sequence    *)
   nmlngth = 10;     (* number of characters in species name    *)
   maxtrees = 100;   (* maximum number of tied trees stored     *)
   maxuser = 10;     (* maximum number of user-defined trees *)

TYPE steparray = ARRAY[1..maxsites] OF INTEGER;
   aas = (ala, arg, asn, asp, cys, gln, glu, gly, his, ileu, leu, lys,
          met, phe, pro, ser1, ser2, thr, trp, tyr, val, del, stop,
          asx, glx, ser, unk);
   aaset = SET OF ala..stop;
   sitearray = ARRAY[0..3] OF aaset;
   nodeptr = ^node;   (* nodes will form a binary tree *)
   node = RECORD      (* describes a tip species or an ancestor *)
      ancestor, leftdesc, rtdesc : nodeptr;  (* pointers to nodes *)
      index : INTEGER;      (* number of the node *)
      tip : BOOLEAN;        (* present species are tips of tree *)
      seq : ARRAY[1..maxsites] OF aas; (* the sequence *)
      siteset : sitearray;  (* temporary storage for aa's *)
      maybe, nonzero : BOOLEAN; (* used by printtrav *)
      xcoord, ycoord : INTEGER    (* used by printree *)
      END;
   pointarray = ARRAY[1..maxsz] OF nodeptr;
   longer = ARRAY[0..5] OF INTEGER;

VAR root : nodeptr;
   infile, treefile : TEXT;
   spp, nonodes, chars, inseed, outgrno : INTEGER;
     (* spp = number of species
        nonodes = number of nodes in tree
        chars = number of sites in actual sequences
        outgrno indicates outgroup *)
   jumble, usertree, weights, thresh, trout, outgropt : BOOLEAN;
   weight : steparray;
   treenode : pointarray;      (* pointers to all nodes in tree *)
   name : ARRAY [1..maxsp, 1..nmlngth] OF CHAR; (* names of species *)
   threshold : REAL;
   threshwt : ARRAY [1..maxsites] OF REAL;
   seed : longer;
   enterorder : ARRAY[1..maxsp] OF INTEGER;

   FUNCTION randum (VAR seed : longer) : REAL;
      (* random number generator -- slow but machine independent *)
      VAR i, j, k, sum : INTEGER;
         mult, newseed : longer;
         x : REAL;
   BEGIN (* randum *)
      mult[0] := 13;
      mult[1] := 24;
      mult[2] := 22;
      mult[3] :=  6;
      FOR i := 0 TO 5 DO
         newseed[i] := 0;
      FOR i := 0 TO 5 DO BEGIN
         sum := newseed[i];
         k := i;
         IF i > 3 THEN k := 3;
         FOR j := 0 TO k DO
            sum := sum + mult[j] * seed[i-j];
         newseed[i] := sum;
         FOR j := i TO 4 DO BEGIN
            newseed[j+1] := newseed[j+1] + newseed[j] DIV 64;
            newseed[j] := newseed[j] MOD 64;
            END;
         END;
      seed := newseed;
      seed[5] := seed[5] MOD 4;
      x := 0.0;
      FOR i := 0 TO 5 DO
         x := x / 64.0 + seed[i];
      x := x / 4.0;
      randum := x;
   END; (* randum *)

   FUNCTION digit (ch : CHAR) : BOOLEAN;
   BEGIN (* digit *)
      digit := (ch >= '0') AND (ch <= '9');
   END; (* digit *)

   FUNCTION letter (ch : CHAR) : BOOLEAN;
      (* tests whether ch is a letter
         -- works for both ASCII and EBCDIC codes *)
   BEGIN (* letter *)
      letter := ((ch >= 'A') AND (ch <= 'I'))
                OR ((ch >= 'J') AND (ch <= 'R'))
                OR ((ch >= 'S') AND (ch <= 'Z'))
                OR ((ch >= 'a') AND (ch <= 'i'))
                OR ((ch >= 'j') AND (ch <= 'r'))
                OR ((ch >= 's') AND (ch <= 'z'));
   END; (* letter *)

   PROCEDURE uppercase (VAR ch : CHAR);
      (* convert ch to upper case -- either ASCII or EBCDIC *)
   BEGIN (* uppercase *)
      IF letter(ch) AND (('a' > 'A') AND (ch >= 'a')
                     OR  ('a' < 'A') AND (ch < 'A'))
      THEN ch := CHR(ORD(ch) + ORD('A') - ORD('a'));
   END; (* uppercase *)

   PROCEDURE newline (i, j, k : INTEGER);
      (* go to new line if i is a multiple of j, indent k spaces *)
      VAR m : INTEGER;
   BEGIN (* newline *)
      IF (((i-1) MOD j) = 0) AND (i>1)
      THEN BEGIN
         WRITELN(OUTPUT);
         FOR m := 1 TO k DO WRITE(OUTPUT, ' ');
         END;
   END; (* newline *)


   PROCEDURE doinput;
      (* reads the input data *)

      PROCEDURE inputnumbers;
         (* input the numbers of species and of characters *)
      BEGIN (* inputnumbers *)
         WRITELN(OUTPUT);
         WRITE(OUTPUT, ' Protein sequence parsimony algorithm');
         WRITELN(OUTPUT, ' version 3.1');
         WRITELN(OUTPUT);
         READ(infile, spp, chars);
         WRITELN(OUTPUT, ' ',spp:2,' species, ',chars:3,'  sites');
         IF spp > maxsp
         THEN WRITELN(OUTPUT, ' TOO MANY SPECIES: adjust CONSTants');
         IF chars > maxsites
         THEN WRITELN(OUTPUT, ' TOO MANY SITES: adjust CONSTants');
         WRITELN(OUTPUT);
         nonodes := 2 * spp - 1;
      END; (* inputnumbers *)

      PROCEDURE inputoptions;
         (* input the information on the options *)
         VAR ch : CHAR;
            extranum, i : INTEGER;

         PROCEDURE inputhresh;
            (* input the threshold *)
         BEGIN (* inputhresh *)
            READLN(infile, threshold);
            thresh := TRUE;
         END; (* inputhresh *)

         PROCEDURE printhresh;
            (* print out the threshold value *)
         BEGIN (* printhresh *)
            WRITELN(OUTPUT, '    Threshold value = ', threshold:10:3);
            WRITELN(OUTPUT);
         END; (* printhresh *)

         PROCEDURE inputweights;
            (* input the character weights, 0-9 and A-Z for weights
               0 - 35 *)
            VAR ch : CHAR;
               i : INTEGER;
         BEGIN (* inputweights *)
            FOR i := 1 TO nmlngth-1 DO READ(infile, ch);
            FOR i := 1 TO chars DO BEGIN
               REPEAT
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, ch);
               UNTIL ch <> ' ';
               weight[i] := 1;
               IF digit(ch)
               THEN weight[i] := ORD(ch) - ORD('0')
               ELSE IF letter(ch)
                  THEN BEGIN
                     uppercase (ch);
                     IF (ch >= 'A') AND (ch <= 'I')
                     THEN weight[i] := ORD(ch) - ORD('A') + 10
                     ELSE IF (ch >= 'J') AND (ch <= 'R')
                        THEN weight[i] := ORD(ch) - ORD('J') + 19
                        ELSE weight[i] := ORD(ch) - ORD('S') + 28;
                     END
                  ELSE WRITELN(OUTPUT, ' Bad weight character: ',ch);
               END;
            READLN(infile);
            weights := TRUE;
         END; (* inputweights *)

         PROCEDURE printweights;
            (* print out the weights of sites *)
            VAR i, j, k : INTEGER;
         BEGIN (* printweights *)
            WRITELN(OUTPUT,'    Sites are weighted as follows:');
            WRITE(OUTPUT, '        ');
            FOR i := 0 TO 9 DO WRITE(OUTPUT, i:3);
            WRITELN(OUTPUT);
            WRITELN(OUTPUT,'     *---------------------------------');
            FOR j := 0 TO (chars DIV 10) DO BEGIN
               WRITE(OUTPUT, 10*j:5, '!  ');
               FOR i := 0 TO 9 DO BEGIN
                  k := 10*j + i;
                  IF (k>0) AND (k<=chars)
                  THEN WRITE(OUTPUT, weight[k]:3)
                  ELSE WRITE(OUTPUT, '   ');
                  END;
               WRITELN(OUTPUT);
               END;
            WRITELN(OUTPUT);
         END; (* printweights *)

      BEGIN (* inputoptions *)
         jumble := FALSE;
         outgrno := 1;
         outgropt := FALSE;
         thresh := FALSE;
         trout := FALSE;
         usertree := FALSE;
         weights := FALSE;
         extranum := 0;
         WHILE NOT EOLN(infile) DO BEGIN
            READ(infile, ch);
            uppercase (ch);
            IF (ch = 'J') OR (ch = 'O') OR (ch = 'T')
               OR (ch = 'U') OR (ch = 'W') OR (ch = 'Y')
            THEN CASE ch OF
                  'J' : BEGIN
                        jumble := TRUE;
                        extranum := extranum + 1;
                        END;
                  'O' : BEGIN
                        outgropt := TRUE;
                        extranum := extranum + 1;
                        END;
                  'T' : extranum := extranum + 1;
                  'U' : usertree := TRUE;
                  'W' : extranum := extranum + 1;
                  'Y' : trout := TRUE
                  END
            ELSE IF ch <> ' '
               THEN WRITELN(OUTPUT, ' Bad Option Character: ', ch);
            END;
         READLN(infile);
         threshold := spp;
         FOR i := 1 TO chars DO weight[i] := 1;
         FOR i := 1 TO extranum DO BEGIN
            READ(infile, ch);
            uppercase (ch);
            IF ch = 'J'
            THEN READLN(infile, inseed);
            IF ch = 'O'
            THEN READLN(infile, outgrno);
            IF ch = 'T'
            THEN inputhresh;
            IF ch = 'W'
            THEN inputweights;
            END;
         IF jumble
         THEN BEGIN
            WRITELN(OUTPUT, ' Random number seed = ', inseed:8);
            WRITELN(OUTPUT);
            FOR i := 0 TO 5 DO seed[i] := 0;
            i := 0;
            REPEAT
               seed[i] := inseed MOD 64;
               inseed := inseed DIV 64;
               i := i + 1;
            UNTIL inseed = 0;
            END;
         IF thresh
         THEN printhresh;
         IF weights
         THEN printweights
         ELSE FOR i := 1 TO chars DO
            threshwt[i] := threshold * weight[i];
      END; (* inputoptions *)

      PROCEDURE inputdata;
         (* input the names and sequences for each species *)
         VAR i, j : INTEGER;
            aa : aas; (* temporary amino acid for input *)
            charstate : CHAR;  (* character temp'y storage for input *)
      BEGIN (* inputdata *)
         WRITELN(OUTPUT);
         j := nmlngth - 5 + chars DIV 2;
         IF j < nmlngth - 1 THEN j := nmlngth - 1;
         IF j > 39 THEN j := 39;
         WRITE(OUTPUT, ' Name');
         FOR i := 1 TO j DO WRITE(OUTPUT, ' ');
         WRITELN(OUTPUT, 'Sequences');
         WRITE(OUTPUT, ' ----');
         FOR i := 1 TO j DO WRITE(OUTPUT, ' ');
         WRITELN(OUTPUT, '---------');
         WRITELN(OUTPUT);
         FOR i := 1 TO nonodes DO BEGIN
            NEW (treenode[i]);
            treenode[i]^.leftdesc := NIL;
            treenode[i]^.rtdesc := NIL;
            treenode[i]^.ancestor := NIL;
            treenode[i]^.index := i;
            IF i <= spp
            THEN BEGIN
               treenode[i]^.tip := TRUE;
               FOR j := 1 TO nmlngth DO READ(infile, name[i,j]);
               WRITE(OUTPUT, ' ');
               FOR j := 1 TO nmlngth DO WRITE(OUTPUT, name[i,j]);
               WRITE(OUTPUT, '   ');
               FOR j := 1 TO chars DO BEGIN
                  REPEAT
                     IF EOLN(infile) THEN READLN(infile);
                     READ(infile, charstate);
                  UNTIL charstate <> ' ';
                  uppercase (charstate);
                  IF (NOT letter (charstate)) AND
                     (NOT ((charstate = '.') OR (charstate = '?')
                      OR (charstate = '-') OR (charstate = '*')))
                     OR (charstate = 'J') OR (charstate = 'O')
                     OR (charstate = 'U')
                  THEN WRITELN(OUTPUT, ' Warning -- ',
                        'bad amino acid: ', charstate, 
                        ' at position ', j:5, ' of species ', i:3);
                  newline (j, 60, nmlngth+4);
                  WRITE(OUTPUT, charstate);
                  IF ((j MOD 10) = 0) AND ((j MOD 60) <> 0)
                  THEN WRITE(OUTPUT, ' ');
                  CASE charstate OF
                     'A' : aa := ala;
                     'B' : aa := asx;
                     'C' : aa := cys;
                     'D' : aa := asp;
                     'E' : aa := glu;
                     'F' : aa := phe;
                     'G' : aa := gly;
                     'H' : aa := his;
                     'I' : aa := ileu;
                     'K' : aa := lys;
                     'L' : aa := leu;
                     'M' : aa := met;
                     'N' : aa := asn;
                     'P' : aa := pro;
                     'Q' : aa := gln;
                     'R' : aa := arg;
                     'S' : aa := ser;
                     'T' : aa := thr;
                     'V' : aa := val;
                     'W' : aa := trp;
                     'X' : aa := unk;
                     'Y' : aa := tyr;
                     'Z' : aa := glx;
                     '.' : aa := stop;
                     '*' : aa := stop;
                     '?' : aa := unk;
                     '-' : aa := del
                     END;
                  treenode[i]^.seq[j] := aa;
                  END;
               READLN(infile);
               WRITELN(OUTPUT);
               END
            ELSE treenode[i]^.tip := FALSE;
            END;
         WRITELN(OUTPUT);
         WRITELN(OUTPUT);
      END; (* inputdata *)

   BEGIN (* doinput *)
      inputnumbers;
      inputoptions;
      inputdata;
   END; (* doinput *)

   PROCEDURE maketree;
      (* constructs a binary tree from the pointers in treenode.
         adds each node at location which yields highest "likelihood"
         then rearranges the tree for greatest "likelihood" *)
      TYPE placearray = ARRAY[1..maxsz] OF INTEGER;
      VAR i, j, k, numtrees, nextree, which, minwhich : INTEGER;
         like, bestyet, bestlike, gotlike, 
            minsteps, sum, sum2, sd : REAL;
         lastrearr : BOOLEAN;
         nsteps : ARRAY [1..maxtrees] OF REAL;
         fsteps : ARRAY [1..maxuser] OF ARRAY [1..maxsites] OF REAL;
         there, item, nufork, dummy : nodeptr;
         numsteps : steparray;
         place : placearray;
         bestrees : ARRAY[1..maxtrees] OF placearray;
         translate : ARRAY [ala..unk] OF sitearray;

      PROCEDURE setup;
         (* set up set table to get aasets from aas *)
         VAR a, b : aas;
            s, fulldel, fullset : aaset;
      BEGIN (* setup *)
         FOR a := ala TO stop DO
            translate [a][0] := [a];
         translate [ala][1] := [asp, glu, gly, ser1, pro, thr, val];
         translate [arg][1] := [cys, gln, gly, his, ileu, leu, lys,
                                met, pro, ser2, thr, trp, stop];
         translate [asn][1] := [asp, his, ileu, lys, ser2, thr, tyr];
         translate [asp][1] := [ala, asn, glu, gly, his, tyr, val];
         translate [cys][1] := [arg, gly, phe, ser1, ser2, trp, tyr,
                                stop];
         translate [gln][1] := [arg, glu, his, leu, lys, pro, stop];
         translate [glu][1] := [ala, asp, gln, gly, lys, val, stop];
         translate [gly][1] := [ala, arg, asp, cys, glu, ser2,
                                trp, val, stop];
         translate [his][1] := [arg, asn, asp, gln, leu, pro, tyr];
         translate [ileu][1] := [arg, asn, leu, lys, met, phe, ser2,
                                 thr, val];
         translate [leu][1] := [arg, gln, his, ileu, met, phe,
                                pro, ser1, trp, val, stop];
         translate [lys][1] := [arg, asn, gln, glu, ileu, met, thr,
                                stop];
         translate [met][1] := [arg, ileu, leu, lys, val, thr];
         translate [phe][1] := [cys, ileu, leu, ser1, tyr, val];
         translate [pro][1] := [ala, arg, gln, his, leu, ser1, thr];
         translate [ser1][1] := [ala, cys, leu, phe, pro, thr, trp,
                                 tyr, stop];
         translate [ser2][1] := [arg, asn, cys, gly, ileu, thr];
         translate [thr][1] := [ala, arg, asn, ileu, lys, met,
                                pro, ser1, ser2];
         translate [trp][1] := [arg, cys, gly, leu, ser1, stop];
         translate [tyr][1] := [asn, asp, cys, his, phe, ser1, stop];
         translate [val][1] := [ala, asp, glu, gly, ileu,
                                leu, met, phe];
         translate [stop][1] := [arg, cys, gln, glu, gly, leu, lys,
                                 ser1, trp, tyr];
         translate [del][1] := [];
         fulldel := [ala..stop];
         fullset := fulldel - [del];
         translate [unk][0] := fullset;
         translate [asx][0] := [asn, asp];
         translate [glx][0] := [gln, glu];
         translate [ser][0] := [ser1, ser2];
         translate [unk][1] := [];
         translate [asx][1] := translate [asn][1]
                                 + translate [asp][1] - [asn, asp];
         translate [glx][1] := translate [gln][1]
                                 + translate [glu][1] - [gln, glu];
         translate [ser][1] := translate [ser1][1]
                                + translate [ser2][1];
         FOR a := ala TO unk DO BEGIN
            s := [];
            FOR b := ala TO stop DO
               IF b IN translate [a][1]
               THEN s := s + translate [b][1];
            translate [a][2] := (s - translate [a][1])
                                    - translate [a][0];
            translate [a][3] := ((fulldel - translate [a][1])
                                - translate [a][2]) - translate [a][0];
            END;
      END; (* setup *)

      PROCEDURE add (below, newtip, newfork : nodeptr);
         (* inserts the nodes newfork and its left descendant, newtip,
            to the tree.  below becomes newfork's right descendant *)
      BEGIN (* add *)
         newfork^.ancestor := below^.ancestor;
         newfork^.rtdesc := below;
         newfork^.leftdesc := newtip;
         below^.ancestor := newfork;
         newtip^.ancestor := newfork;
         IF newfork^.ancestor <> NIL
         THEN IF newfork^.ancestor^.leftdesc = below
            THEN newfork^.ancestor^.leftdesc := newfork
            ELSE newfork^.ancestor^.rtdesc := newfork;
      END; (* add *)

      PROCEDURE remove (VAR item, fork : nodeptr);
         (* removes nodes item and its ancestor, fork, from the tree.
            the new descendant of fork's ancestor is made to be
            fork's second descendant (other than item).  Also
            returns pointers to the deleted nodes, item and fork *)
         VAR other, below : nodeptr;
      BEGIN (* remove *)
         IF item^.ancestor = NIL
         THEN fork := NIL
         ELSE BEGIN
            fork := item^.ancestor;
            IF item = fork^.leftdesc
            THEN other := fork^.rtdesc
            ELSE other := fork^.leftdesc;
            below := fork^.ancestor;
            other^.ancestor := below;
            IF below <> NIL
            THEN BEGIN
               IF below^.leftdesc = fork
               THEN below^.leftdesc := other
               ELSE below^.rtdesc := other;
               fork^.ancestor := NIL;
               END;
            fork^.leftdesc := NIL;
            fork^.rtdesc := NIL;
            item^.ancestor := NIL;
            END;
      END; (* remove *)

      PROCEDURE evaluate (r: nodeptr);
         (* determines the number of steps needed for a tree. this is
            the minimum number of steps needed to evolve sequences on
            this tree *)
         VAR i, m, steps : INTEGER;
            sum, term : REAL;

         PROCEDURE postorder (p : nodeptr);
            (* traverses a binary tree, calling PROCEDURE fillin at a
               node's descendants before calling fillin at the node *)

            PROCEDURE fillin (q: nodeptr);
               (* sets up for each node in the tree the aa set for site m
                  at that point and counts the changes.  The program
                  spends much of its time in this PROCEDURE *)
               VAR counted : BOOLEAN;
                  aa : aas;
                  s : aaset;
                  l, r : nodeptr;
                  i, j, k, n : INTEGER;
            BEGIN (* fillin *)
               counted := FALSE;
               l := q^.leftdesc;
               r := q^.rtdesc;
               k := 0;
               FOR i := 0 TO 6 DO
                  IF k < 4
                  THEN BEGIN
                     s := [];
                     IF i > 3
                     THEN n := i-3
                     ELSE n := 0;
                     FOR j := n TO i-n DO
                        s := s + (l^.siteset[j] * r^.siteset[i-j]);
                     q^.siteset[k] := s;
                     IF counted OR (s <> [])
                     THEN BEGIN
                        k := k + 1;
                        counted := TRUE;
                        END
                     ELSE IF NOT counted
                        THEN numsteps[m] := numsteps[m] + weight[m];
                     END;
               FOR i := 2 DOWNTO 0 DO
                  FOR aa := ala TO stop DO
                     IF aa IN q^.siteset[i]
                     THEN FOR j := i+1 TO 3 DO
                        q^.siteset[j] := q^.siteset[j]
                                          + translate [aa][j-i];
               s := [];
               FOR i := 0 TO 3 DO BEGIN
                  q^.siteset[i] := q^.siteset[i] - s;
                  s := s + q^.siteset[i];
                  END;
            END; (* fillin *)

         BEGIN (* postorder *)
            IF NOT p^.tip
            THEN BEGIN
               postorder (p^.leftdesc);
               postorder (p^.rtdesc);
               fillin (p);
               END;
         END; (* postorder *)

      BEGIN (* evaluate *)
         FOR m := 1 TO chars DO BEGIN
            FOR i := 1 TO spp DO
               treenode[i]^.siteset := translate [treenode[i]^.seq[m]];
            numsteps[m] := 0;
            postorder (r);
            END;
         sum := 0;
         FOR i := 1 TO chars DO BEGIN
            steps := numsteps[i];
            IF steps <= threshwt[i]
            THEN term := steps
            ELSE term := threshwt[i];
            sum := sum + term;
            IF usertree
            THEN fsteps[which][i] := term;
            END;
         IF usertree
         THEN BEGIN
            nsteps[which] := sum;
            IF which = 1
            THEN BEGIN
               minwhich := 1;
               minsteps := sum;
               END
            ELSE IF sum < minsteps
               THEN BEGIN
                  minwhich := which;
                  minsteps := sum;
                  END;
            END;
         like := -sum;
      END; (* evaluate *)

      PROCEDURE reroot (outgroup : nodeptr);
         (* reorients tree, putting outgroup in desired position.
            Written by Kent Fiala *)
         VAR vertex, oldvertex, ante, otherbranch : nodeptr;
            left : BOOLEAN;
      BEGIN (* reroot *)
         IF NOT (outgroup^.ancestor = root)
         THEN BEGIN
            oldvertex := outgroup;
            vertex := outgroup^.ancestor;
            ante := vertex^.ancestor;
            IF vertex^.rtdesc = outgroup
            THEN BEGIN
               vertex^.rtdesc := vertex^.leftdesc;
               vertex^.leftdesc := outgroup;
               END;
            REPEAT
               IF oldvertex = vertex^.leftdesc
               THEN BEGIN
                  vertex^.leftdesc := ante;
                  left := TRUE;
                  END
               ELSE BEGIN
                  vertex^.rtdesc := ante;
                  left := FALSE;
                  END;
               vertex^.ancestor := oldvertex;
               oldvertex := vertex;
               vertex := ante;
               ante := vertex^.ancestor
            UNTIL vertex = root;
            IF root^.leftdesc = oldvertex
            THEN otherbranch := root^.rtdesc
            ELSE otherbranch := root^.leftdesc;
            IF left
            THEN oldvertex^.leftdesc := otherbranch
            ELSE oldvertex^.rtdesc := otherbranch;
            otherbranch^.ancestor := oldvertex;
            root^.rtdesc := outgroup^.ancestor;
            root^.leftdesc := outgroup;
            outgroup^.ancestor^.ancestor := root;
            outgroup^.ancestor := root;
            END;
      END; (* reroot *)

      PROCEDURE addpreorder (p, item, nufork : nodeptr);
         (* traverses a binary tree, calling PROCEDURE tryadd
            at a node before calling tryadd at its descendants *)
         TYPE knowed = RECORD    (* temporary storage for tree shape *)
            ancestor, leftdesc, rtdesc : nodeptr;
            END;
         VAR treeknowed : ARRAY [1..maxsz] OF knowed;

         PROCEDURE tryadd (p: nodeptr);
            (* temporarily adds one fork and one tip to the tree.
               if the location where they are added yields greater
               "likelihood" than other locations tested up to that
               time, then keeps that location as there *)
            VAR i, pos : INTEGER;
               found : BOOLEAN;
               rute : nodeptr;

            PROCEDURE savetree;
               (* record in place where each species has to be
                  added to reconstruct this tree *)
               VAR i, j : INTEGER;
                  p : nodeptr;
                  done : BOOLEAN;
            BEGIN (* savetree *)
               reroot (treenode[outgrno]);
               FOR i := 1 TO nonodes DO place[i] := 0;
               place[root^.index] := 1;
               FOR i := 1 TO spp DO BEGIN
                  p := treenode[i];
                  WHILE (place[p^.index] = 0) DO BEGIN
                     place[p^.index] := i;
                     p := p^.ancestor;
                     END;
                  IF i > 1
                  THEN BEGIN
                     place[i] := place[p^.index];
                     j := place[p^.index];
                     done := FALSE;
                     WHILE NOT done DO BEGIN
                        place[p^.index] := spp + i - 1;
                        p := p^.ancestor;
                        done := (p = NIL);
                        IF NOT done
                        THEN done := (place[p^.index] <> j);
                        END;
                     END;
                  END;
            END; (* savetree *)

            PROCEDURE findtree;
               (* finds tree given by ARRAY place in ARRAY
                  bestrees by binary search *)
               VAR i, lower, upper : INTEGER;
                  below : BOOLEAN;
            BEGIN (* findtree *)
               below := FALSE;
               lower := 1;
               upper := nextree - 1;
               found := FALSE;
               WHILE (NOT found) AND (lower <= upper) DO BEGIN
                  pos := (lower + upper) DIV 2;
                  i := 3;
                  WHILE (place[i] = bestrees[pos][i]) AND (i <= spp) DO
                     i := i + 1;
                  found := (i > spp);
                  below := place[i] < bestrees[pos][i];
                  IF NOT found
                  THEN BEGIN
                     IF below
                     THEN upper := pos - 1
                     ELSE lower := pos + 1;
                     END;
                  END;
               IF (NOT found) AND (NOT below)
               THEN pos := pos + 1;
            END; (* findtree *)

            PROCEDURE addtree;
               (* puts tree from ARRAY place in its proper position
                  in ARRAY bestrees *)
               VAR i : INTEGER;
            BEGIN (* addtree *)
               FOR i := (nextree - 1) DOWNTO pos DO
                  bestrees[i+1] := bestrees[i];
               bestrees[pos] := place;
               nextree := nextree + 1;
            END; (* addtree *)

         BEGIN (* tryadd *)
            add (p, item, nufork);
            IF root^.ancestor <> NIL THEN root := root^.ancestor;
            evaluate (root);
            IF lastrearr
            THEN BEGIN
               IF like >= bestlike
               THEN BEGIN
                  FOR i := 1 TO nonodes DO BEGIN
                     treeknowed[i].ancestor := treenode[i]^.ancestor;
                     treeknowed[i].leftdesc := treenode[i]^.leftdesc;;
                     treeknowed[i].rtdesc := treenode[i]^.rtdesc;
                     END;
                  rute := root;
                  savetree;
                  FOR i := 1 TO nonodes DO BEGIN
                     treenode[i]^.ancestor := treeknowed[i].ancestor;
                     treenode[i]^.leftdesc := treeknowed[i].leftdesc;;
                     treenode[i]^.rtdesc := treeknowed[i].rtdesc;
                     END;
                  root := rute;
                  IF like > bestlike
                  THEN BEGIN
                     bestlike := like;
                     pos := 1;
                     nextree := 1;
                     addtree;
                     END
                  ELSE BEGIN
                     pos := 0;
                     findtree;
                     IF NOT found
                     THEN IF nextree <= maxtrees
                        THEN addtree;
                     END;
                  END;
               END;
            IF like > bestyet
            THEN BEGIN
               bestyet := like;
               there := p;
               END;
            remove (item, nufork);
            IF nufork = root THEN root := p;
         END; (* tryadd *)

      BEGIN (* addpreorder *)
         IF p <> NIL
         THEN BEGIN
            tryadd (p);
            addpreorder (p^.leftdesc, item, nufork);
            addpreorder (p^.rtdesc, item, nufork);
            END;
      END; (* addpreorder *)

      PROCEDURE rearrange (VAR r: nodeptr);
         (* traverses the tree (preorder), finding any local
            rearrangement which decreases the number of steps.
            if traversal succeeds in increasing the tree's
            "likelihood", PROCEDURE rearrange runs traversal again *)
         VAR success : BOOLEAN;

         PROCEDURE tryrearr (p : nodeptr);
            (* evaluates one rearrangement of the tree.
               if the new tree has greater "likelihood" than the old
               one sets success := TRUE and keeps the new tree.
               otherwise, restores the old tree *)
            VAR frombelow, whereto, forknode : nodeptr;
               oldlike : REAL;
         BEGIN (* tryrearr *)
            IF p^.ancestor <> NIL
            THEN BEGIN
               IF p^.ancestor^.ancestor <> NIL
               THEN BEGIN
                  oldlike := bestyet;
                  IF p^.ancestor^.leftdesc = p
                  THEN frombelow := p^.ancestor^.rtdesc
                  ELSE frombelow := p^.ancestor^.leftdesc;
                  whereto := p^.ancestor^.ancestor;
                  forknode := p^.ancestor;
                  remove (p, forknode);
                  add (whereto, p, forknode);
                  IF r^.ancestor <> NIL THEN r := r^.ancestor;
                  evaluate (r);
                  IF like <= oldlike
                  THEN BEGIN
                     remove (p, forknode);
                     add (frombelow, p, forknode);
                     IF r^.ancestor <> NIL THEN r := r^.ancestor;
                     END
                  ELSE BEGIN
                     success := TRUE;
                     bestyet := like;
                     END;
                  END;
               END;
         END; (* tryrearr *)

         PROCEDURE repreorder (p : nodeptr);
            (* traverses a binary tree, calling PROCEDURE tryrearr
               at a node before calling tryrearr at its descendants *)
         BEGIN (* repreorder *)
            IF p <> NIL
            THEN BEGIN
               tryrearr (p);
               repreorder (p^.leftdesc);
               repreorder (p^.rtdesc);
               END;
         END; (* repreorder *)

      BEGIN (* rearrange *)
         success := TRUE;
         WHILE success DO BEGIN
            success := FALSE;
            repreorder (r);
            END;
      END; (* rearrange *)

      PROCEDURE treeread;
         (* read in user-defined tree and set it up *)
         VAR ch : CHAR;
            nextnode : INTEGER;

         PROCEDURE findch (c : CHAR);
            (* scan forward until find character c *)
         BEGIN (* findch *)
            IF ch <> c
            THEN REPEAT
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, ch);
               UNTIL ch = c;
         END; (* findch *)

         PROCEDURE addelement (VAR p : nodeptr);
            (* recursive procedure adds nodes to user-defined tree *)
            VAR q : nodeptr;
               i, n : INTEGER;
               found : BOOLEAN;
               str : ARRAY[1..nmlngth] OF CHAR;
         BEGIN (* addelement *)
            REPEAT
               IF EOLN(infile) THEN READLN(infile);
               READ(infile, ch);
            UNTIL ch <> ' ';
            IF ch = '('
            THEN BEGIN
               nextnode := nextnode + 1;
               q := treenode[nextnode];
               addelement (q^.leftdesc);
               q^.leftdesc^.ancestor := q;
               findch (',');
               addelement (q^.rtdesc);
               q^.rtdesc^.ancestor := q;
               findch (')');
               p := q;
               END
            ELSE BEGIN
               FOR i := 1 TO nmlngth DO str[i] := ' ';
               n := 1;
               REPEAT
                  IF (ch = '_') THEN ch := ' ';
                  str[n] := ch;
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, ch);
                  n := n + 1;
               UNTIL ((ch = ',') OR (ch = ')') OR (n > nmlngth));
               n := 1;
               REPEAT
                  found := TRUE;
                  FOR i := 1 TO nmlngth DO
                     found := found AND (str[i] = name[n, i]);
                  IF found
                  THEN p := treenode[n]
                  ELSE n := n + 1;
               UNTIL (n > spp) OR found;
               IF n > spp
               THEN BEGIN
                  WRITE(OUTPUT, ' Cannot find species: ');
                  FOR i := 1 TO nmlngth DO
                     WRITE(OUTPUT, str[i]);
                  WRITELN(OUTPUT);
                  END;
               END;
         END; (* addelement *)

      BEGIN (* treeread *)
         root := treenode[spp + 1];
         nextnode := spp;
         root^.ancestor := NIL;
         addelement (root);
         READLN(infile);
      END; (* treeread *)

      PROCEDURE printree;
         (* prints out diagram of the tree *)
         CONST down = 2;
            over = 3;
         VAR tipx, tipy, i, rover : INTEGER;

         PROCEDURE coordinates (p : nodeptr);
            (* establishes coordinates of nodes *)
         BEGIN (* coordinates *)
            IF p^.tip
            THEN BEGIN
               p^.xcoord := tipx;
               p^.ycoord := tipy;
               tipx := tipx + rover;
               tipy := tipy + down;
               END
            ELSE BEGIN
               coordinates (p^.leftdesc);
               p^.xcoord := p^.leftdesc^.xcoord;
               coordinates (p^.rtdesc);
               p^.ycoord := p^.rtdesc^.ycoord;
               END;
         END; (* coordinates *)

         PROCEDURE drawline (i : INTEGER);
            (* draws one row of the tree diagram by moving up tree *)
            VAR p, q : nodeptr;
               n, j : INTEGER;
         BEGIN (* drawline *)
            p := root;
            WRITE(OUTPUT, '   ');
            REPEAT
               IF p^.leftdesc^.ycoord >= i
               THEN p := p^.leftdesc
               ELSE BEGIN
                  q := p^.rtdesc;
                  n := q^.xcoord - p^.xcoord;
                  IF p^.ycoord = i
                  THEN BEGIN
                     IF rover < over
                     THEN BEGIN
                        IF (i = (tipy-down)) AND (p^.xcoord = 0)
                        THEN WRITE(OUTPUT, 'L')
                        ELSE WRITE(OUTPUT, '*');
                        n := n + 1;
                        END
                     ELSE BEGIN
                        IF p^.index - spp >= 10
                        THEN WRITE(OUTPUT, (p^.index - spp):2)
                        ELSE WRITE(OUTPUT, (p^.index - spp):1, '-');
                        END;
                     FOR j := 1 TO n-2 DO WRITE(OUTPUT, '-');
                     END
                  ELSE BEGIN
                     WRITE(OUTPUT, '!');
                     FOR j := 1 TO n-1 DO WRITE(OUTPUT, ' ');
                     END;
                  p := q;
                  END;
            UNTIL p^.tip;
            IF p^.ycoord = i
            THEN FOR j := 1 TO nmlngth DO
                       WRITE(OUTPUT, name[p^.index, j]);
            WRITELN(OUTPUT);
         END; (* drawline *)

      BEGIN (* printree *)
         tipx := 0;
         tipy := 1;
         rover := 100 DIV spp;
         IF rover > over THEN rover := over;
         coordinates (root);
         WRITELN(OUTPUT);
         WRITELN(OUTPUT);
         WRITELN(OUTPUT);
         FOR i := 1 TO tipy-down DO drawline (i);
         WRITE(OUTPUT, '   remember:');
         IF outgropt
         THEN WRITE(OUTPUT, ' (although rooted by outgroup)');
         WRITELN(OUTPUT, ' this is an unrooted tree!');
         WRITELN(OUTPUT);
      END; (* printree *)

      PROCEDURE describe;
         (* prints ancestors, steps and table of numbers of steps in
            each site *)
         VAR i, j, k : INTEGER;

         PROCEDURE hypstates;
            (* fill in and describe states at interior nodes *)
            VAR i : INTEGER;
               bottom : BOOLEAN;
               nothing : sitearray;

            PROCEDURE ancestset (a, b, c : sitearray;
                                    VAR d : sitearray);
               (* sets up the aa set array. *)
               VAR aa : aas;
                  s : aaset;
                  i, j, k, m, n : INTEGER;
                  counted : BOOLEAN;
            BEGIN (* ancestset *)
               counted := FALSE;
               k := 0;
               FOR i := 0 TO 6 DO
                  IF k < 4
                  THEN BEGIN
                     s := [];
                     IF i > 3
                     THEN n := i-3
                     ELSE n := 0;
                     FOR j := n TO i-n DO
                        FOR m := n TO  i-j-n DO
                           s := s + (a[j] * b[m] * c[i-j-m]);
                     d[k] := s;
                     IF counted OR (s <> [])
                     THEN BEGIN
                        k := k + 1;
                        counted := TRUE;
                        END;
                     END;
               FOR i := 2 DOWNTO 0 DO
                  FOR aa := ala TO stop DO
                     IF aa IN d[i]
                     THEN FOR j := i+1 TO 3 DO
                        d[j] := d[j] + translate [aa][j-i];
               s := [];
               FOR i := 0 TO 3 DO BEGIN
                  d[i] := d[i] - s;
                  s := s + d[i];
                  END;
            END; (* ancestset *)

            PROCEDURE firstrav (r : nodeptr; m : INTEGER);
               (* preliminary filling in of ancestral node sets *)
            BEGIN (* firstrav *)
               IF m = 1
               THEN BEGIN
                  r^.maybe := FALSE;
                  r^.nonzero := FALSE;
                  END;
               IF NOT r^.tip
               THEN BEGIN
                  firstrav (r^.leftdesc, m);
                  firstrav (r^.rtdesc, m);
                  ancestset (r^.leftdesc^.siteset, r^.rtdesc^.siteset,
                                nothing, r^.siteset);
                  END
               ELSE r^.siteset := translate[r^.seq[m]];
            END; (* firstrav *)

            PROCEDURE hyptrav (r : nodeptr; VAR hypset : sitearray);
               (*  compute states at one interior node *)
               VAR j : INTEGER;
                  aa : aas;
                  anc, hset : aaset;
                  temparray, ancset : sitearray;
            BEGIN (* hyptrav *)
               IF NOT r^.tip
               THEN BEGIN
                  ancestset (hypset, r^.leftdesc^.siteset,
                             r^.rtdesc^.siteset, temparray);
                  r^.siteset := temparray;
                  END;
               IF NOT bottom
               THEN anc := r^.ancestor^.siteset[0];
               IF NOT r^.tip
               THEN BEGIN
                  j := 0;
                  WHILE r^.siteset[j] = [] DO j := j + 1;
                  hset := r^.siteset[j];
                  r^.seq[i] := unk;
                  FOR aa := ala TO stop DO
                     IF hset = [aa]
                     THEN r^.seq[i] := aa;
                  IF hset = [asn, asp]
                  THEN r^.seq[i] := asx;
                  IF hset = [gln, gly]
                  THEN r^.seq[i] := glx;
                  IF hset = [ser1, ser2]
                  THEN r^.seq[i] := ser;
                  END;
               bottom := FALSE;
               r^.nonzero := r^.nonzero OR ((r^.siteset[0]
                                              * anc) = []);
               r^.maybe := r^.maybe OR (r^.siteset[0] <> anc);
               IF NOT r^.tip
               THEN BEGIN
                  temparray := r^.leftdesc^.siteset;
                  ancestset(hypset, r^.rtdesc^.siteset,
                            r^.leftdesc^.siteset, ancset);
                  hyptrav (r^.leftdesc, ancset);
                  ancestset(hypset, temparray,
                            r^.rtdesc^.siteset, ancset);
                  hyptrav (r^.rtdesc, ancset);
                  END;
            END; (* hyptrav *)

            PROCEDURE printtrav (q : nodeptr);
               (* traverse nodes, working out sequence at each *)
               VAR i : INTEGER;

               PROCEDURE hyprint;
                  (* print out state symbols at node *)
                  VAR dot : BOOLEAN;
                     ch : CHAR;
                     aa, anc : aas;
               BEGIN (* hyprint *)
                  dot := FALSE;
                  aa := q^.seq[i];
                  CASE aa OF
                     ala : ch := 'A';
                     asx : ch := 'B';
                     cys : ch := 'C';
                     asp : ch := 'D';
                     glu : ch := 'E';
                     phe : ch := 'F';
                     gly : ch := 'G';
                     his : ch := 'H';
                     ileu : ch := 'I';
                     lys : ch := 'K';
                     leu : ch := 'L';
                     met : ch := 'M';
                     asn : ch := 'N';
                     pro : ch := 'P';
                     gln : ch := 'Q';
                     arg : ch := 'R';
                     ser : ch := 'S';
                     ser1 : ch := 'S';
                     ser2 : ch := 'S';
                     thr : ch := 'T';
                     val : ch := 'V';
                     trp : ch := 'W';
                     tyr : ch := 'Y';
                     glx : ch := 'Z';
                     del : ch := '-';
                     stop : ch := '*';
                     unk : ch := '?'
                     END;
                  anc := unk;
                  IF NOT bottom
                  THEN anc := q^.ancestor^.seq[i];
                  dot := (aa = anc) AND (aa <> unk) AND (NOT bottom);
                  IF dot
                  THEN WRITE(OUTPUT, '.')
                     ELSE WRITE(OUTPUT, ch);
                  IF (i MOD 10) = 0
                  THEN WRITE(OUTPUT, ' ');
               END; (* hyprint *)

            BEGIN (* printtrav *)
               IF bottom
               THEN BEGIN
                  IF NOT outgropt
                  THEN WRITE(OUTPUT, '        ')
                  ELSE WRITE(OUTPUT, ' root   ');
                  END
               ELSE WRITE(OUTPUT,
                    (q^.ancestor^.index - spp):4, '    ');
               IF q^.tip
               THEN FOR i := 1 TO nmlngth DO
                  WRITE(OUTPUT, name[q^.index, i])
               ELSE WRITE(OUTPUT, (q^.index - spp):4, '      ');
               IF bottom
               THEN WRITE(OUTPUT, '          ')
               ELSE IF q^.nonzero
                  THEN WRITE(OUTPUT, '   yes    ')
                  ELSE IF q^.maybe
                     THEN WRITE(OUTPUT, '  maybe   ')
                     ELSE WRITE(OUTPUT, '   no     ');
               FOR i := 1 TO chars DO BEGIN
                  newline (i, 40, nmlngth+18);
                  hyprint;
                  END;
               WRITELN(OUTPUT);
               bottom := FALSE;
               IF NOT q^.tip
               THEN BEGIN
                  printtrav (q^.leftdesc);
                  printtrav (q^.rtdesc);
                  END;
            END; (* printtrav *)

         BEGIN (* hypstates *)
            WRITELN(OUTPUT);
            WRITE(OUTPUT, ' From    To     Any Steps?    ');
            WRITELN(OUTPUT, 'State at upper node');
            WRITE(OUTPUT, '                             ');
            WRITELN(OUTPUT, ' ( . means same as in the node',
                                    ' below it on tree)');
            WRITELN(OUTPUT);
            nothing := translate[unk];
            FOR i := 1 TO chars DO BEGIN
               firstrav (root, i);
               bottom := TRUE;
               hyptrav (root, nothing);
               END;
            bottom := TRUE;
            printtrav (root);
         END; (* hypstates *)

         PROCEDURE treeout (p : nodeptr);
            (* write out file with representation of final tree *)
            VAR i, n : INTEGER;
               c : CHAR;
         BEGIN (* treeout *)
            IF p^.tip
            THEN BEGIN
               n := 0;
               FOR i := 1 TO nmlngth DO
                  IF name[p^.index, i] <> ' '
                  THEN n := i;
               FOR i := 1 TO n DO BEGIN
                  c := name[p^.index, i];
                  IF c = ' '
                  THEN c := '_';
                  WRITE(treefile, c);
                  END;
               END
            ELSE BEGIN
               WRITE(treefile, '(');
               treeout (p^.leftdesc);
               WRITE(treefile, ',');
               treeout (p^.rtdesc);
               WRITE(treefile, ')');
               END;
            IF p = root
            THEN WRITELN(treefile, ';');
         END; (* treeout *)

      BEGIN (* describe *)
         WRITELN(OUTPUT);
         WRITELN(OUTPUT, ' requires a total of ',-like:10:3);
         WRITELN(OUTPUT);
         IF weights THEN WRITE(OUTPUT, ' weighted');
         WRITELN(OUTPUT, ' steps in each position:');
         WRITE(OUTPUT, '       ');
         FOR i := 0 TO 9 DO WRITE(OUTPUT, i:4);
         WRITELN(OUTPUT);
         WRITE(OUTPUT, '      *------------------------------------');
         WRITELN(OUTPUT, '-----');
         FOR i := 0 TO (chars DIV 10) DO BEGIN
            WRITE(OUTPUT, 10*i:6);
            WRITE(OUTPUT, '!');
            FOR j := 0 TO 9 DO BEGIN
               k := 10*i + j;
               IF (k = 0) OR (k > chars)
               THEN WRITE(OUTPUT, '    ')
               ELSE WRITE(OUTPUT, numsteps[k]:4);
               END;
            WRITELN(OUTPUT);
            END;
         hypstates;
         WRITELN(OUTPUT);
         WRITELN(OUTPUT);
         IF trout
         THEN treeout (root);
      END; (* describe *)

   BEGIN (* maketree *)
      setup;
      IF NOT usertree
      THEN BEGIN
         FOR i := 1 TO spp DO enterorder[i] := i;
         IF jumble
         THEN FOR i := 1 TO spp DO BEGIN
            j := TRUNC(randum(seed)*spp) + 1;
            k := enterorder[j];
            enterorder[j] := enterorder[i];
            enterorder[i] := k;
            END;
         add (treenode[enterorder[2]],
            treenode[enterorder[1]], treenode[spp+1]);
         root := treenode[spp+1];
         WRITELN(OUTPUT, ' Adding species:');
         WRITE(OUTPUT, '    ');
         FOR i := 1 TO nmlngth DO
            WRITE(OUTPUT, name[enterorder[1],i]);
         WRITELN(OUTPUT);
         WRITE(OUTPUT, '    ');
         FOR i := 1 TO nmlngth DO
            WRITE(OUTPUT, name[enterorder[2],i]);
         WRITELN(OUTPUT);
         lastrearr := FALSE;
         FOR i := 3 TO spp DO BEGIN
            bestyet := -32000;
            there := root;
            item := treenode[enterorder[i]];
            nufork := treenode[spp + i - 1];
            addpreorder (root, item, nufork);
            add (there, item, nufork);
            IF root^.ancestor <> NIL THEN root := root^.ancestor;
            like := bestyet;
            rearrange (root);
            WRITE(OUTPUT, '    ');
            FOR j := 1 TO nmlngth DO
               WRITE(OUTPUT, name[enterorder[i],j]);
            WRITELN(OUTPUT);
            lastrearr := (i = spp);
            IF lastrearr
            THEN BEGIN
               WRITELN(OUTPUT, ' Doing global rearrangements');
               bestlike := bestyet;
               nextree := 1;
               REPEAT
                  WRITE(OUTPUT, '    ');
                  gotlike := bestlike;
                  FOR j := 1 TO nonodes DO BEGIN
                     bestyet := -32000;
                     there := root;
                     item := treenode[j];
                     IF item <> root
                     THEN BEGIN
                        nufork := treenode[j]^.ancestor;
                        IF root = nufork
                        THEN BEGIN
                           IF nufork^.leftdesc = item
                           THEN root := nufork^.rtdesc
                           ELSE root := nufork^.leftdesc;
                           END;
                        remove (item, nufork);
                        addpreorder (root, item, nufork);
                        add (there, item, nufork);
                        IF root^.ancestor <> NIL
                        THEN root := root^.ancestor;
                        END;
                     WRITE(OUTPUT, '.');
                     END;
                  WRITELN(OUTPUT);
               UNTIL bestlike <= gotlike;
               END;
            END;
         FOR i := spp DOWNTO 2 DO
            remove (treenode[i], dummy);
         WRITELN(OUTPUT);
         IF (nextree = 2)
         THEN WRITELN(OUTPUT, ' One most parsimonious tree found:')
         ELSE WRITELN(OUTPUT, nextree-1:6, ' trees in all found');
         IF nextree > maxtrees + 1
         THEN BEGIN
            WRITELN(OUTPUT, ' here are the first', maxtrees:4,
                        ' of them');
            nextree := maxtrees + 1;
            END;
         WRITELN(OUTPUT);
         FOR i := 1 TO nextree-1 DO BEGIN
            add (treenode[2], treenode[1], treenode[spp+1]);
            root := treenode[spp+1];
            FOR j := 3 TO spp DO BEGIN
               add(treenode[bestrees[i][j]], treenode[j],
                                              treenode[spp+j-1]);
               IF root^.ancestor <> NIL THEN root := root^.ancestor;
               END;
            reroot (treenode[outgrno]);
            evaluate (root);
            printree;
            describe;
            FOR j := 2 TO spp DO
               remove (treenode[j], dummy);
            END;
         IF trout
         THEN BEGIN
            WRITELN(OUTPUT, ' Trees also written onto file');
            WRITELN(OUTPUT);
            END;
         END
      ELSE BEGIN
         WRITE(OUTPUT, ' User-defined tree');
         IF numtrees > 1
         THEN WRITE(OUTPUT, 's');
         WRITELN(OUTPUT, ':');
         WRITELN(OUTPUT);
         WRITELN(OUTPUT);
         WRITELN(OUTPUT);
         READLN(infile, numtrees);
         FOR which := 1 TO numtrees DO BEGIN
            treeread;
            IF outgropt
            THEN reroot (treenode[outgrno]);
            evaluate (root);
            printree;
            describe;
            END;
         WRITELN(OUTPUT);
         IF (numtrees > 1) AND (chars > 1)
         THEN BEGIN
            WRITE(OUTPUT, ' Tree    Steps     Diff Steps    Its S.D.');
            WRITELN(OUTPUT, '   Significantly worse?');
            WRITELN(OUTPUT);
            FOR which := 1 TO numtrees DO BEGIN
               WRITE(OUTPUT, which:4, nsteps[which]:12:5);
               IF minwhich = which
               THEN WRITELN(OUTPUT, '  <------ best')
               ELSE BEGIN
                  sum := 0.0;
                  sum2 := 0.0;
                  FOR j := 1 TO chars DO
                     IF weight[j] > 0
                     THEN BEGIN
                        sum := sum + weight[j] * (fsteps[which][j]
                                                  -fsteps[minwhich][j]);
                        sum2 := sum2 + weight[j] * SQR(fsteps[which][j]
                                                  -fsteps[minwhich][j]);
                        END;
                  sd := SQRT((chars/(chars-1))*(sum2-SQR(sum/chars)));
                  WRITE(OUTPUT, (nsteps[which]-minsteps):12:5, sd:12:4);
                  IF sum > 1.95996*sd
                  THEN WRITELN(OUTPUT, '           Yes')
                  ELSE WRITELN(OUTPUT, '           No');
                  END;
               END;
            WRITELN(OUTPUT);
            WRITELN(OUTPUT);
            END;
         END;
   END; (* maketree *)

BEGIN (* Protein parsimony by uphill search *)
   (* reads in spp, chars, and the data. Then calls maketree to
      construct the tree *)
   (* ASSIGN(infile, 'INFILE');  -- Turbo Pascal only *)
   RESET(infile);
   doinput;
   IF trout
   THEN BEGIN
      (* ASSIGN(treefile, 'TREEFILE');  -- Turbo Pascal only *)
      REWRITE(treefile);
      END;
   maketree;
(* CLOSE(infile);  -- may need this here *)
   IF trout
   THEN BEGIN
      (* CLOSE(treefile);  -- may need this here *)
      END;
END. (* Protein parsimony by uphill search *)

